"""Module provider for Gehirn"""
import base64
import json
import logging
import re

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["gehirn.jp"]


def provider_parser(subparser):
    """Construct subparser for Gehirn"""
    subparser.add_argument(
        "--auth-token", help="specify access token for authentication"
    )
    subparser.add_argument(
        "--auth-secret", help="specify access secret for authentication"
    )


BUILD_FORMATS = {
    "A": "{address}",
    "AAAA": "{address}",
    "CNAME": "{cname}",
    "TXT": "{data}",
    "NS": "{nsdname}",
    "MX": "{prio} {exchange}",
    "SRV": "{prio} {weight} {port} {target}",
}

FORMAT_RE = {
    "A": re.compile(r"(?P<address>.+)"),
    "AAAA": re.compile(r"(?P<address>.+)"),
    "CNAME": re.compile(r"(?P<cname>.+)"),
    "TXT": re.compile(r"(?P<data>.+)"),
    "NS": re.compile(r"(?P<nsdname>.+)"),
    "MX": re.compile(r"(?P<prio>\d+)\s+(?P<exchange>.+)"),
    "SRV": re.compile(
        r"(?P<prio>\d+)\s+(?P<weight>\d+)\s+(?P<port>\d+)\s+(?P<target>.+)"
    ),
}


class Provider(BaseProvider):
    """Provider class for Gehirn"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.version_id = None
        self.api_endpoint = "https://api.gis.gehirn.jp/dns/v1"

    def _authenticate(self):
        payload = self._get("/zones")

        domains = [item for item in payload if item["name"] == self.domain]
        if not domains:
            raise AuthenticationError("No domain found")

        self.domain_id = domains[0]["id"]
        self.version_id = domains[0]["current_version_id"]

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        name = self._full_name(name)
        a_record = self._parse_content(rtype, content)

        record = None
        records = self._get_records(rtype=rtype, name=name)
        if not records:
            record = {
                "type": rtype,
                "name": name,
                "enable_alias": False,
                "ttl": self._get_lexicon_option("ttl"),
                "records": [],
            }
        else:
            record = records[0]

        if a_record in record["records"]:
            LOGGER.debug("create_record: %s", True)
            return True

        record["records"].append(a_record)
        self._update_internal_record(record)
        LOGGER.debug("create_record: %s", True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        records = []
        if name:
            name = self._full_name(name)
        for record in self._get_records(rtype=rtype, name=name):
            for a_record in record["records"]:
                content = self._build_content(record["type"], a_record)
                processed_record = {
                    "type": record["type"],
                    "name": record["name"].rstrip("."),
                    "ttl": record["ttl"],
                    "content": content,
                    "id": f"{record['id']}.{base64.b64encode(content.encode('utf-8')).decode('ascii')}",
                }
                self._parse_content(record["type"], processed_record["content"])
                records.append(processed_record)

        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)
        return records

    # Create or update a record.
    def _update_record(self, identifier=None, rtype=None, name=None, content=None):

        if name:
            name = self._full_name(name)

        if not identifier:
            if not (rtype and name and content):
                raise Exception("type, name and content must be specified.")

            records = self._get_records(rtype=rtype, name=name)

            if not records:
                self._create_record(rtype=rtype, name=name, content=content)
                LOGGER.debug("update_record: %s", True)
                return True

            record = {
                "id": records[0]["id"],
                "type": rtype,
                "name": name,
                "enable_alias": False,
                "ttl": self._get_lexicon_option("ttl"),
                "records": [self._parse_content(rtype, content)],
            }

        else:
            # with identifier
            records = self._get_records(identifier=identifier)
            if not records:
                raise Exception("Record identifier could not be found.")

            record = records[0]

            if "." in identifier:
                # modify single record
                self._delete_record(identifier=identifier)
                self._create_record(
                    rtype=rtype or record["type"],
                    name=name or record["name"],
                    content=content,
                )
                LOGGER.debug("update_record: %s", True)
                return True
            # update entire record
            if rtype:
                record["type"] = rtype
            if name:
                record["name"] = name
            record["ttl"] = self._get_lexicon_option("ttl")
            if content:
                record["records"] = [self._parse_content(record["type"], content)]

        self._update_internal_record(record)
        LOGGER.debug("update_record: %s", True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier:
            if "." not in identifier:
                # delete entire record
                path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{identifier}"
                self._delete(path)
                LOGGER.debug("delete_record: %s", True)
                return True

            record_identifier = identifier.split(".")[1]

            records = self._get_records(identifier=identifier)
            if not records:
                raise Exception("Record identifier could not be found.")

            record = records[0]
            for index, a_record in enumerate(record["records"]):
                target_content = self._build_content(record["type"], a_record)
                target_identifier = base64.b64encode(
                    target_content.encode("utf-8")
                ).decode("ascii")

                if target_identifier == record_identifier:
                    del record["records"][index]
                if not record["records"]:
                    # delete entire record
                    path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{record['id']}"
                    self._delete(path)
                else:
                    self._update_internal_record(record)

                LOGGER.debug("delete_record: %s", True)
                return True

            raise Exception("Record identifier could not be found.")

        record = None
        if name is not None:
            name = self._full_name(name)
        if content is not None:
            content = self._bind_format_target(rtype, content)
            record = self._parse_content(rtype, content)

        records = self._get_records(rtype=rtype, name=name)

        for a_record in records:
            if record and record in a_record["records"]:
                a_record["records"].remove(record)
                if a_record["records"]:
                    self._update_internal_record(a_record)
                    continue

            path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{a_record['id']}"
            self._delete(path)

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _full_name(self, record_name):
        record_name = super(Provider, self)._full_name(record_name)
        if not record_name.endswith("."):
            record_name += "."
        return record_name

    def _bind_format_target(self, rtype, target):
        if rtype == "CNAME" and not target.endswith("."):
            target += "."
        return target

    def _filter_records(self, records, identifier=None, rtype=None, name=None):
        filtered_records = []

        if identifier:
            identifier = identifier.split(".")[0]

        for record in records:
            if rtype and record["type"] != rtype:
                continue
            if name and record["name"] != name:
                continue
            if identifier and record["id"] != identifier:
                continue
            filtered_records.append(record)
        return filtered_records

    def _get_records(self, identifier=None, rtype=None, name=None):
        path = f"/zones/{self.domain_id}/versions/{self.version_id}/records"
        return self._filter_records(
            self._get(path), identifier=identifier, rtype=rtype, name=name
        )

    def _update_internal_record(self, record):
        if record.get("id"):
            # PUT
            path = f"/zones/{self.domain_id}/versions/{self.version_id}/records/{record['id']}"
            return self._put(path, record)

        # POST
        path = f"/zones/{self.domain_id}/versions/{self.version_id}/records"
        return self._post(path, record)

    def _build_content(self, rtype, record):
        return BUILD_FORMATS[rtype].format(**record)

    def _parse_content(self, rtype, content):
        return FORMAT_RE[rtype].match(content).groupdict()

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        default_auth = HTTPBasicAuth(
            self._get_provider_option("auth_token"),
            self._get_provider_option("auth_secret"),
        )

        query_string = ""
        if query_params:
            query_string = json.dumps(query_params)

        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_string,
            data=json.dumps(data) if data else None,
            headers=default_headers,
            auth=default_auth,
        )
        try:
            # if the request fails for any reason, throw an error.
            response.raise_for_status()
        except BaseException:
            LOGGER.error(response.text)
            raise
        return response.json()



"""Module provider for TransIP"""
import binascii
import json
import logging
import uuid
from base64 import b64decode, b64encode
from typing import Any, Dict, List, Optional

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import load_pem_private_key

try:
    from simplejson import JSONDecodeError
except ImportError:
    from json import JSONDecodeError  # type: ignore[misc]

from lexicon.exceptions import LexiconError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS: List[str] = []

API_BASE_URL = "https://api.transip.nl/v6"


def provider_parser(subparser):
    """Configure provider parser for TransIP"""
    subparser.add_argument(
        "--auth-username", help="specify username for authentication"
    )
    subparser.add_argument(
        "--auth-api-key",
        help="specify the private key to use for API authentication, in PEM format: can be either "
        "the path of the key file (eg. /tmp/key.pem) or the base64 encoded content of this "
        "file prefixed by 'base64::' (eg. base64::eyJhbGciOyJ...)",
    )
    subparser.add_argument(
        "--auth-key-is-global",
        action="store_true",
        help="set this flag is the private key used is a global key with no IP whitelist restriction",
    )


class Provider(BaseProvider):
    """
    Provider class for TransIP

    provider_options can be overwritten by a Provider to setup custom defaults.
    They will be overwritten by any options set via the CLI or Env.
    order is:

    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.provider_name = "transip"
        self.domain_id = None

        private_key_conf = self._get_provider_option("auth_api_key")
        if private_key_conf.startswith("base64::"):
            private_key_bytes = b64decode(private_key_conf.replace("base64::", ""))
        else:
            with open(
                private_key_conf,
                "rb",
            ) as file:
                private_key_bytes = file.read()

        self.private_key = load_pem_private_key(private_key_bytes, password=None)
        self.token: str

    def _authenticate(self):
        request_body = {
            "login": self._get_provider_option("auth_username"),
            "nonce": uuid.uuid4().hex,
            "global_key": self._get_provider_option("auth_key_is_global") or False,
        }

        request_body_bytes = json.dumps(request_body).encode()

        signature = self.private_key.sign(
            request_body_bytes,
            padding.PKCS1v15(),
            hashes.SHA512(),
        )

        headers = {"Signature": b64encode(signature).decode()}

        response = requests.request(
            "POST", f"{API_BASE_URL}/auth", json=request_body, headers=headers
        )
        response.raise_for_status()

        self.token = response.json()["token"]

        data = self._get(f"/domains/{self.domain}")

        self.domain_id = data["domain"]["authCode"]

    def _create_record(self, rtype: str, name: str, content: str) -> bool:
        if not rtype or not name or not content:
            raise Exception(
                "Error, rtype, name and content are mandatory to create a record."
            )

        identifier = Provider._identifier(
            {"type": rtype, "name": self._full_name(name), "content": content}
        )

        if any(
            record
            for record in self._list_records(rtype=rtype, name=name, content=content)
            if record["id"] == identifier
        ):
            LOGGER.debug("create_record (ignored, duplicate): %s", identifier)
            return True

        data = {
            "dnsEntry": {
                "type": rtype,
                "name": self._relative_name(name),
                "content": content,
                "expire": self._get_lexicon_option("ttl"),
            },
        }

        self._post(f"/domains/{self.domain}/dns", data=data)

        LOGGER.debug("create_record: %s", identifier)

        return True

    def _list_records(
        self,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        data = self._get(f"/domains/{self.domain}/dns")

        records = []
        for entry in data["dnsEntries"]:
            record = {
                "type": entry["type"],
                "name": self._full_name(entry["name"]),
                "ttl": entry["expire"],
                "content": entry["content"],
            }
            record["id"] = Provider._identifier(record)
            records.append(record)

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            records = [
                record for record in records if record["name"] == self._full_name(name)
            ]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: %s", records)

        return records

    def _update_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if not identifier and (not rtype or not name):
            raise Exception("Error, identifier or rtype+name parameters are required.")

        if identifier:
            records = self._list_records()
            records_to_update = [
                record for record in records if record["id"] == identifier
            ]
        else:
            records_to_update = self._list_records(rtype=rtype, name=name)

        if not records_to_update:
            raise Exception(
                f"Error, could not find a record for given identifier: {identifier}"
            )

        if len(records_to_update) > 1:
            LOGGER.warning(
                "Warning, multiple records found for given parameters, "
                "only first one will be updated: %s",
                records_to_update,
            )

        record = records_to_update[0]

        # TransIP API is not designed to update one record out of several records
        # matching the same type+name (eg. multi-valued TXT entries).
        # To circumvent the limitation, we remove first the record to update, then
        # recreate it with the updated content.

        data = {
            "dnsEntry": {
                "type": record["type"],
                "name": self._relative_name(record["name"]),
                "content": record["content"],
                "expire": record["ttl"],
            },
        }

        self._request("DELETE", f"/domains/{self.domain}/dns", data=data)

        data["dnsEntry"]["content"] = content

        self._post(f"/domains/{self.domain}/dns", data=data)

        LOGGER.debug("update_record: %s", record["id"])

        return True

    def _delete_record(
        self,
        identifier: Optional[str] = None,
        rtype: Optional[str] = None,
        name: Optional[str] = None,
        content: Optional[str] = None,
    ) -> bool:
        if identifier:
            records = self._list_records()
            records = [record for record in records if record["id"] == identifier]

            if not records:
                raise LexiconError(
                    f"Could not find a record matching the identifier provider: {identifier}"
                )
        else:
            records = self._list_records(rtype, name, content)

        for record in records:
            data = {
                "dnsEntry": {
                    "type": record["type"],
                    "name": self._relative_name(record["name"]),
                    "content": record["content"],
                    "expire": record["ttl"],
                },
            }

            self._request("DELETE", f"/domains/{self.domain}/dns", data=data)

        LOGGER.debug("delete_records: %s %s %s %s", identifier, rtype, name, content)

        return True

    def _request(
        self,
        action: str = "GET",
        url: str = "/",
        data: Optional[Dict] = None,
        query_params: Optional[Dict] = None,
    ) -> Optional[Dict[str, Any]]:
        response = requests.request(
            action,
            f"{API_BASE_URL}{url}",
            params=query_params,
            json=data,
            headers={"Authorization": f"Bearer {self.token}"},
        )

        response.raise_for_status()

        try:
            return response.json()
        except JSONDecodeError:
            return None

    @staticmethod
    def _identifier(record):
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(("type=" + record.get("type", "") + ",").encode("utf-8"))
        digest.update(("name=" + record.get("name", "") + ",").encode("utf-8"))
        digest.update(("content=" + record.get("content", "") + ",").encode("utf-8"))

        return binascii.hexlify(digest.finalize()).decode("utf-8")[0:7]



"""Parsers definition for the Lexicon command-line interface"""
import argparse
import importlib
import os

from lexicon import discovery


def generate_base_provider_parser() -> argparse.ArgumentParser:
    """Function that generates the base provider to be used by all dns providers."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument(
        "action",
        help="specify the action to take",
        default="list",
        choices=["create", "list", "update", "delete"],
    )
    parser.add_argument(
        "domain", help="specify the domain, supports subdomains as well"
    )
    parser.add_argument(
        "type",
        help="specify the entry type",
        default="TXT",
        choices=["A", "AAAA", "CNAME", "MX", "NS", "SOA", "TXT", "SRV", "LOC"],
    )

    parser.add_argument("--name", help="specify the record name")
    parser.add_argument("--content", help="specify the record content")
    parser.add_argument("--ttl", type=int, help="specify the record time-to-live")
    parser.add_argument("--priority", help="specify the record priority")
    parser.add_argument(
        "--identifier", help="specify the record for update or delete actions"
    )
    parser.add_argument(
        "--log_level",
        help="specify the log level",
        default="ERROR",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG", "NOTSET"],
    )
    parser.add_argument(
        "--output",
        help=(
            "specify the type of output: by default a formatted table (TABLE), "
            "a formatted table without header (TABLE-NO-HEADER), "
            "a JSON string (JSON) or no output (QUIET)"
        ),
        default="TABLE",
        choices=["TABLE", "TABLE-NO-HEADER", "JSON", "QUIET"],
    )
    return parser


def generate_cli_main_parser() -> argparse.ArgumentParser:
    """Using all providers available, generate a parser that will be used by Lexicon CLI"""
    parser = argparse.ArgumentParser(
        description="Create, Update, Delete, List DNS entries"
    )

    parser.add_argument(
        "--version",
        help="show the current version of lexicon",
        action="version",
        version=f"%(prog)s {discovery.lexicon_version()}",
    )
    parser.add_argument("--delegated", help="specify the delegated domain")
    parser.add_argument(
        "--config-dir",
        default=os.getcwd(),
        help="specify the directory where to search lexicon.yml and "
        "lexicon_[provider].yml configuration files "
        "(default: current directory).",
    )
    subparsers = parser.add_subparsers(
        dest="provider_name", help="specify the DNS provider to use"
    )
    subparsers.required = True

    for provider, available in discovery.find_providers().items():
        provider_module = importlib.import_module("lexicon.providers." + provider)
        provider_parser = getattr(provider_module, "provider_parser")

        subparser = subparsers.add_parser(
            provider,
            help=f"{provider} provider",
            parents=[generate_base_provider_parser()],
        )
        provider_parser(subparser)

        if not available:
            subparser.epilog = (
                "WARNING: some required dependencies for this provider are not "
                f"installed. Please run `pip install lexicon[{provider}]` first before using it."
            )

    return parser



"""Module provider for Linode V4"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["linode.com"]


def provider_parser(subparser):
    """Configure provider parser for Linode V4"""
    subparser.add_argument("--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Linode V4"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.linode.com/v4/"

    def _authenticate(self):
        self.domain_id = None
        payload = self._get(
            "domains", query_params={"filter": {"domain": self.domain.lower()}}
        )
        if payload["data"]:
            self.domain_id = payload["data"][0]["id"]
        else:
            raise AuthenticationError("Domain not found")

    def _create_record(self, rtype, name, content):
        if not self._list_records(rtype, name, content):
            if name:
                name = self._relative_name(name).lower()

            self._post(
                f"domains/{self.domain_id}/records",
                data={"name": name, "type": rtype, "target": content, "ttl_sec": 0},
            )

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        resources_url = f"domains/{self.domain_id}/records"

        if name:
            name = self._relative_name(name).lower()

        processed_records = []

        payload = self._get(resources_url)
        for page in range(1, payload["pages"] + 1, 1):
            if page > 1:
                payload = self._get(resources_url, query_params={"page": page})

            resource_list = payload["data"]
            if rtype:
                resource_list = [
                    resource for resource in resource_list if resource["type"] == rtype
                ]
            if name:
                resource_list = [
                    resource
                    for resource in resource_list
                    if self._relative_name(resource["name"]).lower() == name
                ]
            if content:
                resource_list = [
                    resource
                    for resource in resource_list
                    if resource["target"] == content
                ]

            for resource in resource_list:
                processed_records.append(
                    {
                        "id": resource["id"],
                        "type": resource["type"],
                        "name": self._full_name(resource["name"]).lower(),
                        "ttl": resource["ttl_sec"],
                        "content": resource["target"],
                    }
                )

        LOGGER.debug("list_records: %s", processed_records)
        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            resources = self._list_records(rtype, name, None)
            identifier = resources[0]["id"] if resources else None

        LOGGER.debug("update_record: %s", identifier)

        if name:
            name = self._relative_name(name).lower()

        url = f"domains/{self.domain_id}/records/{identifier}"
        self._put(
            url,
            data={
                "name": name.lower() if name else None,
                "type": rtype if rtype else None,
                "target": content if content else None,
            },
        )

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_resource_id = []
        if not identifier:
            resources = self._list_records(rtype, name, content)
            delete_resource_id = [resource["id"] for resource in resources]
        else:
            delete_resource_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_resource_id)

        for resource_id in delete_resource_id:
            self._delete(f"domains/{self.domain_id}/records/{resource_id}")

        return True

    # Helpers
    def _request(self, action="GET", url="", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._get_provider_option('auth_token')}",
        }

        request_filter = query_params["filter"] if "filter" in query_params else None
        if request_filter is not None:
            default_headers["X-Filter"] = json.dumps(request_filter)
            del query_params["filter"]

        request_url = f"{self.api_endpoint}{url}"

        response = requests.request(
            action,
            request_url,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == "DELETE":
            return ""
        result = response.json()
        return result



"""Module provider for Linode"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["linode.com"]


def provider_parser(subparser):
    """Module provider for Linode"""
    subparser.add_argument("--auth-token", help="specify api key for authentication")


class Provider(BaseProvider):
    """Provider class for Linode"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.linode.com/api/"

    def _authenticate(self):
        self.domain_id = None
        payload = self._get("domain.list")
        for domain in payload["DATA"]:
            if domain["DOMAIN"].lower() == self.domain.lower():
                self.domain_id = domain["DOMAINID"]
                break
        else:
            raise AuthenticationError("Domain not found")

    def _create_record(self, rtype, name, content):
        if not self._list_records(rtype, name, content):
            self._get(
                "domain.resource.create",
                query_params={
                    "DomainID": self.domain_id,
                    "Name": self._relative_name(name).lower(),
                    "Type": rtype,
                    "Target": content,
                    "TTL_sec": 0,
                },
            )

        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        payload = self._get(
            "domain.resource.list", query_params={"DomainID": self.domain_id}
        )
        resource_list = payload["DATA"]
        if rtype:
            resource_list = [
                resource for resource in resource_list if resource["TYPE"] == rtype
            ]
        if name:
            cmp_name = self._relative_name(name).lower()
            resource_list = [
                resource
                for resource in resource_list
                if resource["NAME"].lower() == cmp_name
            ]
        if content:
            resource_list = [
                resource for resource in resource_list if resource["TARGET"] == content
            ]

        processed_records = []
        for resource in resource_list:
            processed_records.append(
                {
                    "id": resource["RESOURCEID"],
                    "type": resource["TYPE"],
                    "name": self._full_name(resource["NAME"]).lower(),
                    "ttl": resource["TTL_SEC"],
                    "content": resource["TARGET"],
                }
            )
        LOGGER.debug("list_records: %s", processed_records)
        return processed_records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            resources = self._list_records(rtype, name, None)
            identifier = resources[0]["id"] if resources else None

        LOGGER.debug("update_record: %s", identifier)

        self._get(
            "domain.resource.update",
            query_params={
                "DomainID": self.domain_id,
                "ResourceID": identifier,
                "Name": self._relative_name(name).lower() if name else None,
                "Type": rtype if rtype else None,
                "Target": content if content else None,
            },
        )

        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_resource_id = []
        if not identifier:
            resources = self._list_records(rtype, name, content)
            delete_resource_id = [resource["id"] for resource in resources]
        else:
            delete_resource_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_resource_id)

        for resource_id in delete_resource_id:
            self._get(
                "domain.resource.delete",
                query_params={"DomainID": self.domain_id, "ResourceID": resource_id},
            )

        return True

    # Helpers
    def _request(self, action="GET", url="", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }

        query_params["api_key"] = self._get_provider_option("auth_token")
        query_params["resultFormat"] = "JSON"
        query_params["api_action"] = url

        response = requests.request(
            action,
            self.api_endpoint,
            params=query_params,
            data=json.dumps(data),
            headers=default_headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        if action == "DELETE":
            return ""
        result = response.json()
        if result["ERRORARRAY"]:
            raise Exception(f"Linode api error: {result['ERRORARRAY']}")
        return result



"""Module provider for Name.com"""
from __future__ import absolute_import

import logging

from requests import HTTPError, Session
from requests.auth import HTTPBasicAuth

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["name.com"]

DUPLICATE_ERROR = {
    "message": "Invalid Argument",
    "details": "Parameter Value Error - Duplicate Record",
}


def provider_parser(subparser):
    """Configure a subparser for Name.com."""

    subparser.add_argument("--auth-username", help="specify a username")
    subparser.add_argument("--auth-token", help="specify an API token")


class NamecomLoader(
    object
):  # pylint: disable=useless-object-inheritance,too-few-public-methods
    """Loader that handles pagination for the Name.com provider."""

    def __init__(self, get, url, data_key, next_page=1):
        self.get = get
        self.url = url
        self.data_key = data_key
        self.next_page = next_page

    def __iter__(self):
        while self.next_page:
            response = self.get(self.url, {"page": self.next_page})
            for data in response[self.data_key]:
                yield data
            self.next_page = response.get("next_page")


class NamecomProvider(BaseProvider):
    """Provider implementation for Name.com."""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.api_endpoint = "https://api.name.com/v4"
        self.session = Session()

    def _authenticate(self):
        self.session.auth = HTTPBasicAuth(
            username=self._get_provider_option("auth_username"),
            password=self._get_provider_option("auth_token"),
        )

        # checking domain existence
        domain_name = self.domain
        for domain in NamecomLoader(self._get, "/domains", "domains"):
            if domain["domainName"] == domain_name:
                self.domain_id = domain_name
                return

        raise Exception("{} domain does not exist".format(domain_name))

    def _create_record(self, rtype, name, content):
        data = {
            "type": rtype,
            "host": self._relative_name(name),
            "answer": content,
            "ttl": self._get_lexicon_option("ttl"),
        }

        if rtype in ("MX", "SRV"):
            # despite the documentation says a priority is
            # required for MX and SRV, it's actually optional
            priority = self._get_lexicon_option("priority")
            if priority:
                data["priority"] = priority

        url = "/domains/{}/records".format(self.domain)
        try:
            record_id = self._post(url, data)["id"]
        except HTTPError as error:
            response = error.response
            if response.status_code == 400 and response.json() == DUPLICATE_ERROR:
                LOGGER.warning("create_record: duplicate record has been skipped")
                return True
            raise

        LOGGER.debug("create_record: record %s has been created", record_id)

        return record_id

    def _list_records(self, rtype=None, name=None, content=None):
        url = "/domains/{}/records".format(self.domain)
        records = []

        for raw in NamecomLoader(self._get, url, "records"):
            record = {
                "id": raw["id"],
                "type": raw["type"],
                "name": raw["fqdn"][:-1],
                "ttl": raw["ttl"],
                "content": raw["answer"],
            }
            records.append(record)

        LOGGER.debug("list_records: retrieved %s records", len(records))

        if rtype:
            records = [record for record in records if record["type"] == rtype]
        if name:
            name = self._full_name(name)
            records = [record for record in records if record["name"] == name]
        if content:
            records = [record for record in records if record["content"] == content]

        LOGGER.debug("list_records: filtered %s records", len(records))

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            if not (rtype and name):
                raise ValueError("Record identifier or rtype+name must be specified")
            records = self._list_records(rtype, name)
            if not records:
                raise Exception("There is no record to update")

            if len(records) > 1:
                filtered_records = [
                    record for record in records if record["content"] == content
                ]
                if filtered_records:
                    records = filtered_records

                if len(records) > 1:
                    raise Exception(
                        "There are multiple records to update: {}".format(
                            ", ".join(record["id"] for record in records)
                        )
                    )

            record_id = records[0]["id"]
        else:
            record_id = identifier

        data = {"ttl": self._get_lexicon_option("ttl")}

        # even though the documentation says a type and an answer
        # are required, they are not required actually
        if rtype:
            data["type"] = rtype
        if name:
            data["host"] = self._relative_name(name)
        if content:
            data["answer"] = content

        url = "/domains/{}/records/{}".format(self.domain, record_id)
        record_id = self._put(url, data)["id"]
        logging.debug("update_record: record %s has been updated", record_id)

        return record_id

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if not identifier:
            if not (rtype and name):
                raise ValueError("Record identifier or rtype+name must be specified")
            records = self._list_records(rtype, name, content)
            if not records:
                LOGGER.warning("delete_record: there is no record to delete")
                return False
            record_ids = [record["id"] for record in records]
        else:
            record_ids = [
                identifier,
            ]

        for record_id in record_ids:
            url = "/domains/{}/records/{}".format(self.domain, record_id)
            self._delete(url)
            LOGGER.debug("delete_record: record %s has been deleted", record_id)

        return True

    def _get_raw_record(self, record_id):
        url = "/domains/{}/records/{}".format(self.domain, record_id)
        return self._get(url)

    def _request(self, action="GET", url="/", data=None, query_params=None):
        response = self.session.request(
            method=action, url=self.api_endpoint + url, json=data, params=query_params
        )
        response.raise_for_status()
        return response.json()


Provider = NamecomProvider



"""Module provider for Cloudflare"""
import json
import logging

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ["cloudflare.com"]


def provider_parser(subparser):
    """Return the parser for this provider"""
    subparser.description = """
        There are two ways to provide an authentication granting edition to the target CloudFlare DNS zone.
        1 - A Global API key, with --auth-username and --auth-token flags.
        2 - An unscoped API token (permissions Zone:Zone(read) + Zone:DNS(edit) for all zones), with --auth-token flag.
        3 - A scoped API token (permissions Zone:Zone(read) + Zone:DNS(edit) for one zone), with --auth-token and --zone-id flags.
    """
    subparser.add_argument(
        "--auth-username",
        help="specify email address for authentication (for Global API key only)",
    )
    subparser.add_argument(
        "--auth-token",
        help="specify token for authentication (Global API key or API token)",
    )
    subparser.add_argument(
        "--zone-id",
        help="specify the zone id (if set, API token can be scoped to the target zone)",
    )


class Provider(BaseProvider):
    """Provider class for Cloudflare"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = "https://api.cloudflare.com/client/v4"

    def _authenticate(self):
        zone_id = self._get_provider_option("zone_id")
        if not zone_id:
            payload = self._get("/zones", {"name": self.domain, "status": "active"})

            if not payload["result"]:
                raise AuthenticationError("No domain found")
            if len(payload["result"]) > 1:
                raise AuthenticationError(
                    "Too many domains found. This should not happen"
                )

            self.domain_id = payload["result"][0]["id"]
        else:
            payload = self._get(f"/zones/{zone_id}")

            if not payload["result"]:
                raise AuthenticationError(f"No domain found for Zone ID {zone_id}")

            self.domain_id = zone_id

    # Create record. If record already exists with the same content, do nothing'
    def _create_record(self, rtype, name, content):
        content, cf_data = self._format_content(rtype, content)
        data = {
            "type": rtype,
            "name": self._full_name(name),
            "content": content,
            "data": cf_data,
        }
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        payload = {"success": True}
        try:
            payload = self._post(f"/zones/{self.domain_id}/dns_records", data)
        except requests.exceptions.HTTPError as err:
            already_exists = next(
                (
                    True
                    for error in err.response.json()["errors"]
                    if error["code"] == 81057
                ),
                False,
            )
            if not already_exists:
                raise

        LOGGER.debug("create_record: %s", payload["success"])
        return payload["success"]

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def _list_records(self, rtype=None, name=None, content=None):
        filter_obj = {"per_page": 100}
        if rtype:
            filter_obj["type"] = rtype
        if name:
            filter_obj["name"] = self._full_name(name)
        if content:
            filter_obj["content"] = content

        records = []
        while True:
            payload = self._get(f"/zones/{self.domain_id}/dns_records", filter_obj)

            LOGGER.debug("payload: %s", payload)

            for record in payload["result"]:
                processed_record = {
                    "type": record["type"],
                    "name": record["name"],
                    "ttl": record["ttl"],
                    "content": record["content"],
                    "id": record["id"],
                }
                records.append(processed_record)

            pages = payload["result_info"]["total_pages"]
            page = payload["result_info"]["page"]
            if page >= pages:
                break
            filter_obj["page"] = page + 1

        LOGGER.debug("list_records: %s", records)
        LOGGER.debug("Number of records retrieved: %d", len(records))
        return records

    # Create or update a record.
    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if identifier is None:
            records = self._list_records(rtype, name)
            if len(records) == 1:
                identifier = records[0]["id"]
            elif len(records) < 1:
                raise Exception(
                    "No records found matching type and name - won't update"
                )
            else:
                raise Exception(
                    "Multiple records found matching type and name - won't update"
                )

        data = {}
        if rtype:
            data["type"] = rtype
        if name:
            data["name"] = self._full_name(name)
        if content:
            data["content"] = content
        if self._get_lexicon_option("ttl"):
            data["ttl"] = self._get_lexicon_option("ttl")

        payload = self._put(f"/zones/{self.domain_id}/dns_records/{identifier}", data)

        LOGGER.debug("update_record: %s", payload["success"])
        return payload["success"]

    # Delete an existing record.
    # If record does not exist, do nothing.
    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self._list_records(rtype, name, content)
            delete_record_id = [record["id"] for record in records]
        else:
            delete_record_id.append(identifier)

        LOGGER.debug("delete_records: %s", delete_record_id)

        for record_id in delete_record_id:
            self._delete(f"/zones/{self.domain_id}/dns_records/{record_id}")

        LOGGER.debug("delete_record: %s", True)
        return True

    # Helpers
    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        headers = {"Content-Type": "application/json"}
        if self._get_provider_option("auth_username"):
            headers["X-Auth-Email"] = self._get_provider_option("auth_username")
            headers["X-Auth-Key"] = self._get_provider_option("auth_token")
        else:
            headers[
                "Authorization"
            ] = f"Bearer {self._get_provider_option('auth_token')}"
        response = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=json.dumps(data),
            headers=headers,
        )
        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    def _format_content(self, rtype, content):
        """
        Special case handling from some record types that Cloudflare needs
        formatted differently

        Returns new values for the content and data properties to be sent
        on the request
        """
        data = None
        if rtype == "SSHFP":
            # For some reason the CloudFlare API does not let you set content
            # directly when creating an SSHFP record. You need to pass the
            # fields that make up the record seperately, then the API joins
            # them back together
            _fp = content.split(" ")
            data = {"algorithm": _fp[0], "type": _fp[1], "fingerprint": _fp[2]}
            content = None

        return content, data



"""
This unit test suite ensures that the lexicon client works correctly when used as a library.
In particular:
    - relevant provider should be resolved correctly from config,
    - config should be passed correctly to provider,
    - relevant provider method should be invoked for a given config.
"""
import contextlib
import importlib
import pkgutil
from types import ModuleType
from unittest import mock

import pytest

from lexicon.config import ConfigResolver
from lexicon.exceptions import ProviderNotAvailableError
from lexicon.providers.base import Provider as BaseProvider


class Provider(BaseProvider):
    """
    Fake provider to simulate the provider resolution from configuration,
    and to have execution traces when lexicon client is invoked
    """

    def _authenticate(self):
        print("Authenticate action")

    def _create_record(self, rtype, name, content):
        return {
            "action": "create",
            "domain": self.domain,
            "type": rtype,
            "name": name,
            "content": content,
        }

    def _list_records(self, rtype=None, name=None, content=None):
        return {
            "action": "list",
            "domain": self.domain,
            "type": rtype,
            "name": name,
            "content": content,
        }

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        return {
            "action": "update",
            "domain": self.domain,
            "identifier": identifier,
            "type": rtype,
            "name": name,
            "content": content,
        }

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        return {
            "action": "delete",
            "domain": self.domain,
            "identifier": identifier,
            "type": rtype,
            "name": name,
            "content": content,
        }

    def _request(self, action="GET", url="/", data=None, query_params=None):
        # Not use for tests
        pass


@contextlib.contextmanager
def mock_fake_provider():
    """
    Create a fake provider module, and mock relevant
    functions to make it appear as a real module.
    """
    original_iter = pkgutil.iter_modules
    original_import = importlib.import_module
    with mock.patch("lexicon.discovery.pkgutil.iter_modules") as mock_iter, mock.patch(
        "lexicon.client.importlib.import_module"
    ) as mock_import:

        def return_iter(path):
            """
            This will include an adhoc fakeprovider module
            to the normal return of pkgutil.iter_modules.
            """
            modules = list(original_iter(path))
            modules.append((None, "fakeprovider", None))
            return modules

        mock_iter.side_effect = return_iter

        def return_import(module_name):
            """
            This will return a adhoc fakeprovider module if necessary,
            or fallback to the normal return of importlib.import_module.
            """
            if module_name == "lexicon.providers.fakeprovider":
                module = ModuleType("lexicon.providers.fakeprovider")
                setattr(module, "Provider", Provider)
                return module
            return original_import(module_name)

        mock_import.side_effect = return_import

        yield


@pytest.fixture
def lexicon_client():
    """Return the lexicon_client"""
    return importlib.import_module("lexicon.client")


@pytest.fixture(autouse=True)
def fake_provider():
    """Activate the fake_provider mock"""
    with mock_fake_provider():
        yield


def test_unknown_provider_raises_error(lexicon_client):
    with pytest.raises(ProviderNotAvailableError):
        lexicon_client.Client(
            ConfigResolver().with_dict(
                {
                    "action": "list",
                    "provider_name": "unknownprovider",
                    "domain": "example.com",
                    "type": "TXT",
                    "name": "fake",
                    "content": "fake",
                }
            )
        )


def test_missing_required_client_config_parameter_raises_error(lexicon_client):
    with pytest.raises(AttributeError):
        lexicon_client.Client(
            ConfigResolver().with_dict(
                {
                    "no-action": "list",
                    "provider_name": "fakeprovider",
                    "domain": "example.com",
                    "type": "TXT",
                    "name": "fake",
                    "content": "fake",
                }
            )
        )
    with pytest.raises(AttributeError):
        lexicon_client.Client(
            ConfigResolver().with_dict(
                {
                    "action": "list",
                    "no-provider_name": "fakeprovider",
                    "domain": "example.com",
                    "type": "TXT",
                    "name": "fake",
                    "content": "fake",
                }
            )
        )
    with pytest.raises(AttributeError):
        lexicon_client.Client(
            ConfigResolver().with_dict(
                {
                    "action": "list",
                    "provider_name": "fakeprovider",
                    "no-domain": "example.com",
                    "type": "TXT",
                    "name": "fake",
                    "content": "fake",
                }
            )
        )
    with pytest.raises(AttributeError):
        lexicon_client.Client(
            ConfigResolver().with_dict(
                {
                    "action": "list",
                    "provider_name": "fakeprovider",
                    "domain": "example.com",
                    "no-type": "TXT",
                    "name": "fake",
                    "content": "fake",
                }
            )
        )


def test_missing_optional_client_config_parameter_does_not_raise_error(lexicon_client):
    lexicon_client.Client(
        ConfigResolver().with_dict(
            {
                "action": "list",
                "provider_name": "fakeprovider",
                "domain": "example.com",
                "type": "TXT",
                "no-name": "fake",
                "no-content": "fake",
            }
        )
    )


def test_list_action_is_correctly_handled_by_provider(capsys, lexicon_client):
    client = lexicon_client.Client(
        ConfigResolver().with_dict(
            {
                "action": "list",
                "provider_name": "fakeprovider",
                "domain": "example.com",
                "type": "TXT",
                "name": "fake",
                "content": "fake-content",
            }
        )
    )
    results = client.execute()

    out, _ = capsys.readouterr()

    assert "Authenticate action" in out
    assert results["action"] == "list"
    assert results["domain"] == "example.com"
    assert results["type"] == "TXT"
    assert results["name"] == "fake"
    assert results["content"] == "fake-content"


def test_create_action_is_correctly_handled_by_provider(capsys, lexicon_client):
    client = lexicon_client.Client(
        ConfigResolver().with_dict(
            {
                "action": "create",
                "provider_name": "fakeprovider",
                "domain": "example.com",
                "type": "TXT",
                "name": "fake",
                "content": "fake-content",
            }
        )
    )
    results = client.execute()

    out, _ = capsys.readouterr()

    assert "Authenticate action" in out
    assert results["action"] == "create"
    assert results["domain"] == "example.com"
    assert results["type"] == "TXT"
    assert results["name"] == "fake"
    assert results["content"] == "fake-content"


def test_update_action_is_correctly_handled_by_provider(capsys, lexicon_client):
    client = lexicon_client.Client(
        ConfigResolver().with_dict(
            {
                "action": "update",
                "provider_name": "fakeprovider",
                "domain": "example.com",
                "identifier": "fake-id",
                "type": "TXT",
                "name": "fake",
                "content": "fake-content",
            }
        )
    )
    results = client.execute()

    out, _ = capsys.readouterr()

    assert "Authenticate action" in out
    assert results["action"] == "update"
    assert results["domain"] == "example.com"
    assert results["identifier"] == "fake-id"
    assert results["type"] == "TXT"
    assert results["name"] == "fake"
    assert results["content"] == "fake-content"


def test_delete_action_is_correctly_handled_by_provider(capsys, lexicon_client):
    client = lexicon_client.Client(
        ConfigResolver().with_dict(
            {
                "action": "delete",
                "provider_name": "fakeprovider",
                "domain": "example.com",
                "identifier": "fake-id",
                "type": "TXT",
                "name": "fake",
                "content": "fake-content",
            }
        )
    )
    results = client.execute()

    out, _ = capsys.readouterr()

    assert "Authenticate action" in out
    assert results["action"] == "delete"
    assert results["domain"] == "example.com"
    assert results["identifier"] == "fake-id"
    assert results["type"] == "TXT"
    assert results["name"] == "fake"
    assert results["content"] == "fake-content"



"""
Lexicon Zonomi and Rimuhosting Provider

Author: Juan Rossi, 2017

Zonomi API Docs: https://zonomi.com/app/dns/dyndns.jsp
Rimuhosting API Docs: https://rimuhosting.com/dns/dyndns.jsp

Implementation notes:
* Lots of tricks taken from the PowerDNS API
* The Zonomi API does not assign a unique identifier to each record in the way
  that Lexicon expects. We work around this by creating an ID based on the record
  name, type and content, which when taken together are always unique
* The  API has no notion of 'create a single record' or 'delete a single
  record'. All operations are either 'replace the RRSet with this new set of records'
  or 'delete all records for this name and type. Similarly, there is no notion of
  'change the content of this record', because records are identified by their name,
  type and content.
"""
import logging
from xml.etree import ElementTree

import requests

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

APIENTRYPOINT = {
    "zonomi": "https://zonomi.com/app",
    "rimuhosting": "https://rimuhosting.com",
}

NAMESERVER_DOMAINS = ["zonomi.com"]


def provider_parser(subparser):
    """Configure provider parser for Zonomi"""
    subparser.add_argument("--auth-token", help="specify token for authentication")
    subparser.add_argument(
        "--auth-entrypoint",
        help="use Zonomi or Rimuhosting API",
        choices=["zonomi", "rimuhosting"],
    )


class Provider(BaseProvider):
    """Provider class for Zonomi"""

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        if not self._get_provider_option("auth_token"):
            raise Exception("Error, application key is not defined")

        self.api_endpoint = APIENTRYPOINT.get("zonomi")

        if self._get_provider_option("auth_entrypoint"):
            self.api_endpoint = APIENTRYPOINT.get(
                self._get_provider_option("auth_entrypoint")
            )

    def _authenticate(self):

        payload = self._get(
            "/dns/dyndns.jsp",
            {"action": "QUERY", "name": "**." + self.domain, "type": "SOA"},
        )

        if payload.find("is_ok").text != "OK:":
            raise AuthenticationError(f"Error with api {payload.find('is_ok').text}")

        self.domain_id = self.domain

    def _make_identifier(self, rtype, name, content):
        return f"{rtype}/{self._full_name(name)}={content}"

    def _parse_identifier(self, identifier):
        parts = identifier.split("/")
        rtype = parts[0]
        parts = parts[1].split("=")
        name = parts[0]
        content = "=".join(parts[1:])
        return rtype, name, content

    def _create_record(self, rtype, name, content):
        request = {
            "action": "SET",
            "type": rtype,
            "name": self.domain,
            "value": content,
        }

        if name is not None:
            request["name"] = self._full_name(name)

        if self._get_lexicon_option("ttl"):
            request["ttl"] = self._get_lexicon_option("ttl")

        if self._get_lexicon_option("priority"):
            request["prio"] = self._get_lexicon_option("priority")

        payload = self._get("/dns/dyndns.jsp", request)

        if payload.find("is_ok").text != "OK:":
            raise Exception(f"An error occurred: {payload.find('is_ok').text}")

        LOGGER.debug("create_record: %s", True)
        return True

    def _list_records(self, rtype=None, name=None, content=None):
        records = []

        request = {"action": "QUERY", "name": "**." + self.domain}

        if rtype is not None:
            request["type"] = rtype
        if name is not None:
            request["name"] = self._full_name(name)
        if content is not None:
            request["value"] = content

        payload = self._get("/dns/dyndns.jsp", request)
        for rxml in payload.iter("record"):
            processed_record = {
                "type": rxml.attrib["type"],
                "name": rxml.attrib["name"],
                "content": rxml.attrib["content"],
                "id": self._make_identifier(
                    rxml.attrib["type"], rxml.attrib["name"], rxml.attrib["content"]
                ),
                "ttl": rxml.attrib["ttl"].split()[0],
            }
            records.append(processed_record)
        LOGGER.debug("list_records: %s", records)
        return records

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        if identifier is not None:
            rtype, name, content = self._parse_identifier(identifier)

        request = {"action": "DELETE", "name": self.domain}

        if rtype is not None:
            request["type"] = rtype
        if name is not None:
            request["name"] = self._full_name(name)
        if content is not None:
            request["value"] = content

        payload = self._get("/dns/dyndns.jsp", request)

        if payload.find("is_ok").text != "OK:":
            raise Exception(f"An error occurred: {payload.find('is_ok').text}")

        LOGGER.debug("delete_record: %s", True)
        return True

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        self._delete_record(identifier)
        ttype, tname, tcontent = self._parse_identifier(identifier)
        request = {
            "action": "SET",
            "type": ttype,
            "name": self._full_name(tname),
            "value": tcontent,
        }

        if rtype is not None:
            request["type"] = rtype
        if name is not None:
            request["name"] = self._full_name(name)
        if content is not None:
            request["value"] = content
        if self._get_lexicon_option("ttl"):
            request["ttl"] = self._get_lexicon_option("ttl")
        if self._get_lexicon_option("priority"):
            request["prio"] = self._get_lexicon_option("priority")

        payload = self._get("/dns/dyndns.jsp", request)

        if payload.find("is_ok").text != "OK:":
            raise Exception(f"An error occurred: {payload.find('is_ok').text}")

        LOGGER.debug("update_record: %s", True)
        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        else:
            query_params["api_key"] = self._get_provider_option("auth_token")

        response = requests.request(
            action, self.api_endpoint + url, params=query_params
        )
        tree = ElementTree.ElementTree(ElementTree.fromstring(response.content))
        root = tree.getroot()
        if root.tag == "error":
            raise Exception(f"An error occurred: {root.text}")
        response.raise_for_status()
        return root



"""Module provider for OnApp"""
import json
import logging
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from lexicon.exceptions import AuthenticationError
from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS: List[str] = []


def provider_parser(subparser):
    """Configure provider parser for OnApp"""
    subparser.description = """
        The OnApp provider requires your OnApp account\'s email address and
        API token, which can be found on your /profile page on the Control Panel interface.
        The server is your dashboard URL, with format like https://dashboard.youronapphost.org"""
    subparser.add_argument(
        "--auth-username", help="specify email address of the OnApp account"
    )
    subparser.add_argument("--auth-token", help="specify API Key for the OnApp account")
    subparser.add_argument(
        "--auth-server", help="specify URL to the OnApp Control Panel Server"
    )


class Provider(BaseProvider):
    """Provider class for OnApp"""

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.domain_id = None

        if not self._get_provider_option("auth_username"):
            raise Exception("Error, OnApp Email Address is not defined")
        if not self._get_provider_option("auth_token"):
            raise Exception("Error, OnApp API Key is not defined")
        if not self._get_provider_option("auth_server"):
            raise Exception("Error, OnApp Control Panel URL is not defined")

        self.session = requests.Session()

    def _authenticate(self):
        domain = self.domain

        zones = self._get("/dns_zones.json")
        for zone in zones:
            if zone["dns_zone"]["name"] == domain:
                self.domain_id = zone["dns_zone"]["id"]
                break
        else:
            raise AuthenticationError(f"Could not find {domain} in OnApp DNS Zones")

    def _create_record(self, rtype, name, content):
        data = {
            "name": self._relative_name(name),
            "type": rtype,
            self._key_for_record_type(rtype): content,
        }

        ttl = self._get_lexicon_option("ttl")
        if ttl:
            data["ttl"] = f"{ttl}"

        result = self._post(
            f"/dns_zones/{self.domain_id}/records.json", {"dns_record": data}
        )
        LOGGER.debug("create_record: %s", result)

        return True

    def _list_records(self, rtype=None, name=None, content=None):
        records = []

        response = self._get(f"/dns_zones/{self.domain_id}/records.json")
        for record_type in response["dns_zone"]["records"]:

            # For now we do not support other RR types so we ignore them, also see
            # _key_for_record_type
            if record_type not in ("A", "AAAA", "CNAME", "TXT"):
                continue

            if rtype and record_type != rtype:
                continue

            for record in response["dns_zone"]["records"][record_type]:
                record = record["dns_record"]

                if name and record["name"] != self._relative_name(name):
                    continue

                record_content = record[self._key_for_record_type(record_type)]
                if content and record_content != content:
                    continue

                records.append(
                    {
                        "id": record["id"],
                        "name": self._full_name(record["name"]),
                        "type": record["type"],
                        "ttl": record["ttl"],
                        "content": record_content,
                    }
                )

        LOGGER.debug("list_records: %s", records)

        return records

    def _update_record(self, identifier, rtype=None, name=None, content=None):
        if not identifier:
            existing = self._guess_record(rtype, name)
            identifier = existing["id"]

        ttl = self._get_lexicon_option("ttl")

        if not name or not ttl:
            if not existing:
                existing = self._get(
                    f"/dns_zones/{self.domain_id}/records/{identifier}.json"
                )
            if not name:
                name = existing["name"]
            if not ttl:
                ttl = existing["ttl"]

        request = {
            "name": self._relative_name(name),
            "ttl": f"{ttl}",
            self._key_for_record_type(rtype): content,
        }

        result = self._put(
            f"/dns_zones/{self.domain_id}/records/{identifier}.json",
            {"dns_record": request},
        )
        LOGGER.debug("update_record: %s", result)

        return True

    def _delete_record(self, identifier=None, rtype=None, name=None, content=None):
        deletion_ids = []

        if not identifier:
            records = self._list_records(rtype, name, content)
            deletion_ids = [record["id"] for record in records]
        else:
            deletion_ids.append(identifier)

        for one_id in deletion_ids:
            self._delete(f"/dns_zones/{self.domain_id}/records/{one_id}.json")

        LOGGER.debug("delete_record: %s", True)

        return True

    def _request(self, action="GET", url="/", data=None, query_params=None):
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        target = self._get_provider_option("auth_server") + url

        body = ""
        if data is not None:
            body = json.dumps(data)

        auth = HTTPBasicAuth(
            self._get_provider_option("auth_username"),
            self._get_provider_option("auth_token"),
        )

        request = requests.Request(
            action, target, data=body, headers=headers, params=query_params, auth=auth
        )
        prepared_request = self.session.prepare_request(request)

        result = self.session.send(prepared_request)
        result.raise_for_status()

        if result.text:
            return result.json()
        return None

    def _key_for_record_type(self, record_type):
        if record_type in ("A", "AAAA"):
            return "ip"
        if record_type == "CNAME":
            return "hostname"
        if record_type == "TXT":
            return "txt"
        if record_type in ("MX", "NS", "SOA", "SRV", "LOC"):
            raise Exception(
                f"{record_type} record type is not supported in the OnApp Provider"
            )
        raise Exception(f"{record_type} record type is unknown")

    def _guess_record(self, rtype, name=None, content=None):
        records = self._list_records(rtype=rtype, name=name, content=content)
        if len(records) == 1:
            return records[0]
        if len(records) > 1:
            raise Exception(
                "Identifier was not provided and several existing records "
                "match the request for {0}/{1}".format(rtype, name)
            )
        raise Exception(
            "Identifier was not provided and no existing records "
            "match the request for {0}/{1}".format(rtype, name)
        )
